class Conference < ApplicationRecord
  has_many :sessions
  has_many :speakers
  has_many :stages
  has_many :revision_sets
  has_many :import_histories

  serialize :data, coder: JSON

  validates :time_zone, presence: true, inclusion: { in: ActiveSupport::TimeZone.all.map(&:name) }

  has_many :relevant_stage_links, class_name: "RelevantStage"
  has_many :relevant_stages, through: :relevant_stage_links, source: :stage

  def import_in_progress?
    last_import_started_at.present? &&
    (last_import_completed_at.nil? || last_import_started_at > last_import_completed_at) &&
    last_import_error.nil? &&
    last_import_started_at > 30.minutes.ago
  end

  def import_completed?
    last_import_completed_at.present? &&
    (!last_import_started_at.present? || last_import_completed_at >= last_import_started_at)
  end

  def import_failed?
    last_import_error.present? ||
    (last_import_started_at.present? &&
     last_import_completed_at.nil? &&
     last_import_started_at < 30.minutes.ago)
  end

  def import_status
    if import_in_progress?
      "in_progress"
    elsif import_failed?
      "failed"
    elsif import_completed?
      "completed"
    else
      "pending"
    end
  end

  def import_error_summary
    return nil unless last_import_error.present?
    last_import_error.split("\n").first
  end

  def should_retry_import?
    return false unless import_failed?
    return false if retry_count >= 3
    return false if permanent_error?

    last_retry = last_import_started_at || Time.at(0)
    backoff_time = [ 30, 60, 120 ][retry_count] # 30s, 1m, 2m backoff
    Time.current > (last_retry + backoff_time.minutes)
  end

  def permanent_error?
    return false unless last_import_error.present?

    permanent_error_patterns = [
      /Invalid credentials/i,
      /Access denied/i,
      /Not found/i,
      /Invalid URL/i,
      /No such file or directory/i
    ]

    permanent_error_patterns.any? { |pattern| last_import_error.match?(pattern) }
  end

  def retry_count
    import_histories.where("created_at > ?", 24.hours.ago)
                    .where(status: "failed")
                    .count
  end

  def days
    (starts_at.to_date..ends_at.to_date)
  end

  def to_param
    slug
  end

  def starts_at_in_local_time
    starts_at.in_time_zone(time_zone || "UTC")
  end

  def ends_at_in_local_time
    ends_at.in_time_zone(time_zone || "UTC")
  end

  def schedule_url
    data["schedule_url"]
  end

  def filedrop_url
    data["filedrop_url"]
  end

  def engelsystem_url
    data["engelsystem_url"]
  end

  def heartbeat_url
    data["heartbeat_url"]
  end

  # Get the required data fields for the import job class
  def required_data_fields
    return [] if import_job_class.blank?

    begin
      # Dynamically query the import job class for its required fields
      klass = import_job_class.constantize
      klass.respond_to?(:required_data_fields) ? klass.required_data_fields : []
    rescue NameError => e
      Rails.logger.error("Could not find import job class #{import_job_class}: #{e}")
      []
    end
  end

  def fetch_translation_angel_id
    fetch_engelsystem("angeltypes")
      &.find { |t| t["name"] == "Translation Angel" }
      &.dig("id")
  end

  def fetch_engelsystem(endpoint)
    endpoint_url = engelsystem_url + endpoint
    Rails.logger.debug("Querying engelsystem API at #{endpoint_url}")
    response = HTTParty.get(
      endpoint_url,
      headers: {
        "Accept" => "application/json",
        "x-api-key" => fetch_credential("engelsystem_token")
      },
      timeout: 10
    )
    response.success? ? JSON.parse(response.body)["data"] : nil
  rescue StandardError => e
    Rails.logger.warn("Engelsystem query for #{endpoint} failed: #{e.message}")
    nil
  end

  def compare_engelsystem_shifts(additional_conferences = [])
    return unless translation_angel_id = fetch_translation_angel_id
    return unless data = fetch_engelsystem("angeltypes/#{translation_angel_id}/shifts")

    engelsystem_shifts = data.each_with_object({}) do |shift, hash|
      hash[shift["id"]] = shift
                          &.dig("needed_angel_types")
                          &.find { |t| t["angel_type"]["id"] == translation_angel_id }
                          &.dig("entries")
                          &.map { |t| t["user"]["name"] }
    end

    Session
      .where(conference: [ self, *additional_conferences ])
      .where.not(engelsystem_id: nil)
      .includes(assignments: :user)
      .group_by(&:engelsystem_id)
      .each do |engelsystem_id, sessions|
      engelsystem_assigned = engelsystem_shifts[engelsystem_id]
      local_assigned = sessions.flat_map(&:assignments).map { |a| a.user.name }

      only_engelsystem = engelsystem_assigned - local_assigned
      only_local = local_assigned - engelsystem_assigned

      next if only_engelsystem.blank? and only_local.blank?

      puts "============================="
      puts "Session: #{sessions[0].title} (#{engelsystem_id})"
      puts "============================="
      puts "Not signed up in engelsystem: #{only_local.join(', ')}" unless only_local.blank?
      puts "Missing in local assignments: #{only_engelsystem.join(', ')}" unless only_engelsystem.blank?
      puts
    end

    true
  end
end
